import { _decorator, Component, Node } from 'cc';
import * as cc from 'cc';
const { ccclass, property } = _decorator;

@ccclass('Spine')
export class Spine extends Component {


    start() {

        this.addSpine();
    }

    async addSpine() {
        // 创建一个新节点
        let newNode = new cc.Node('dynamicSpine');
        // 将节点添加到当前节点下
        this.node.addChild(newNode);
        
        // 添加spine组件
        let spineComp = newNode.addComponent(cc.sp.Skeleton);
        
        // 设置spine资源
        let skeletonData = new cc.sp.SkeletonData();
        
        try {
            // 使用Promise方式加载远程资源
            let root = "https://cdn-homeland.flash8f.com/pandaHomeLand/app/web_cn/test_file/girl_walk2/"
            const skeletonJsonAsset = await this.loadRemoteText(root + 'girl_front.json');
            const atlasTextAsset = await this.loadRemoteText(root + 'girl_front.atlas');
            const textureAsset = await this.loadRemoteAsset(root + 'girl_front.png');
            var dd = new cc.Texture2D();
            dd.image = textureAsset;
            
            skeletonData.skeletonJson = skeletonJsonAsset;
            skeletonData.atlasText = atlasTextAsset.toString();
            skeletonData.textures = [dd];
            skeletonData.textureNames = ['girl_front.png']; //必须设置正确的图片文件名，否则app里面会崩溃
            skeletonData._uuid = "aaaa"; // 可以传入任意字符串，但不能为空
            
            spineComp.skeletonData = skeletonData;
            spineComp.setSkin("skin/skin_yellow");
            
            // 设置动画
            spineComp.setAnimation(0, "stand", true);
            
            // 设置位置
            newNode.setPosition(100, 0);
        } catch (error) {
            cc.error("加载Spine资源失败:", error);
        }
    }
    
    // 辅助方法：加载远程资源并返回Promise
    private loadRemoteAsset(url: string): Promise<any> {
        return new Promise((resolve, reject) => {
            cc.assetManager.loadRemote(url, (err, asset) => {
                if (err) {
                    reject(err);
                } else {
                    resolve(asset);
                }
            });
        });
    }
    private loadRemoteText(url: string): Promise<any> {
        return new Promise((resolve, reject) => {
            cc.assetManager.loadRemote(url, {ext: '.txt'}, (err, asset:cc.TextAsset) => {
                if (err) {
                    reject(err);
                } else {
                    resolve(asset.text);
                }
            });
        });
    }

}


